//
// -*- C++ -*-
//
//  This file is part of WEEVILGEN.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: sienaclient.cc,v 1.1.1.1 2004/11/02 23:36:43 ywang Exp $
//
// implementation of a random, communicating siena client
#include <cassert>
#include <cstring>
#include <string>
#include <weevil/weevil.h>
//TOF: with this it is dependent from SFF (Siena Fast Forwarding)
#include <siena/types.h>
#include "simple_fwd_types.h"

using namespace ssim;
using namespace std;
using namespace weevil;
using namespace siena;

/**
 * application-specific function and classes
 */
unsigned int get_rand( unsigned int a, unsigned int b )
{
  // We use the following formula to scale rand() to
  // the range of values we are looking for a -> b inclusive:
  // x = a + int( (rand() / (RAND_MAX+1.0))*(b-a+1) )
  return( a + (unsigned int)(((double)rand()/(RAND_MAX+1.0))*(double)(b-a+1)) );
}

class neighborhood{
public:
  std::string id;
  //std::string eventtype;
  //int signaldelay;
  //double probability;
};

/**
 * Event class
 */
class Trigger : public Event
{
public:
  std::string	from;
  std::string	type;
  int		index;

  Trigger( std::string f, std::string t, int i) : from( f ), type ( t ), index( i ) {}
};


/**
 * Event class to signal messages
 */
class MsgTrigger : public Event
{
public:
  std::string	from;
  simple_message	msg;
  string msg_string;
  int		index;

  MsgTrigger( std::string f, simple_message msg, string msg_string, int index) : from( f ), msg( msg ), msg_string(msg_string), index(index) {}
};

/**
 * reactive process class
 */
class SienaClient : public weevil::WeevilProcess
{
  /**
   * inputted properties from m4 configuration
   */
  //parameters for subscription/notification generation
  unsigned int	m_constr_min;
  unsigned int	m_constr_max;
  unsigned int	m_attr_min;
  unsigned int	m_attr_max;
  char*		m_attr_dist_f;
  char*		m_constr_dist_f;
  char*		m_types_dist_f;
  char*		m_op_dist_f;
  char*		m_op_string_dist_f;
  char*		m_str_dist_f;
  char*		m_int_dist_f;
  char*		m_bool_dist_f;
  //number of subscriptions
  unsigned int	m_num_subs;
  //interval between two subscriptions
  Time		m_sub_timeout;
  //interval between a subscription and its unsubscription
  Time		m_unsub_timeout;
  //number of notifications
  unsigned int	m_num_pubs;
  //interval between two notifications
  Time		m_pub_timeout;
  //the list of clients to trigger when subscribe
  vector<neighborhood> m_sub_neighbors;
  //the list of clients to trigger when publish
  //vector<neighborhood*> m_pub_neighbors;
  //the predicate used to test message matching
  vector<simple_filter>	 filters;
  vector<string> filter_strings;
  //the generated messages
  vector<simple_message> messages;
  // number of matching messages
  //int matches;

  /**
   * data structure to be used
   */
  int	        m_sub_index;
  int	        m_pub_index;
  unsigned int	m_left_subs;
  unsigned int	m_left_pubs;
  //Message_Generator defined in gen_message.h
  Message_Generator* message_generator;

public:
  /**
   * reactive process
   */
  SienaClient(char* id);

  virtual void init();
  //virtual void stop();
  virtual void process_event( const Event* event );

  /**
   * assignment of process properties
   */
  void constr_min( unsigned int cmin )
  { m_constr_min = cmin; }

  void constr_max( unsigned int cmax )
  { m_constr_max = cmax; }

  void attr_min( unsigned int amin )
  { m_attr_min = amin; }

  void attr_max( unsigned int amax )
  { m_attr_max = amax; }

  void attr_dist_f( char* adist )
  { m_attr_dist_f = adist; }

  void constr_dist_f( char* cdist )
  { m_constr_dist_f = cdist; }

  void types_dist_f( char* tdist )
  { m_types_dist_f = tdist; }

  void op_dist_f( char* odist )
  { m_op_dist_f = odist; }

  void op_string_dist_f( char* osdist )
  { m_op_string_dist_f = osdist; }

  void str_dist_f( char*	sdist )
  { m_str_dist_f = sdist; }

  void int_dist_f( char*	idist )
  { m_int_dist_f = idist; }

  void bool_dist_f( char* bdist )
  { m_bool_dist_f = bdist; }

  void num_subs( unsigned int n )
  { m_num_subs = n; m_left_subs = n; }

  void sub_timeout( Time t )
  { m_sub_timeout = t; }

  void unsub_timeout( Time t )
  { m_unsub_timeout = t; }

  void num_pubs( unsigned int n )
  { m_num_pubs = n; m_left_pubs = n;}

  void pub_timeout( Time t )
  { m_pub_timeout = t; }

  /* neighborhood* getobject( std::string neighborstr ){
    int begin = 0;
    int end = neighborstr.find_first_of(" ", begin);
    neighborhood* a_neighbor = new neighborhood();
    a_neighbor->id = neighborstr.substr(begin, end-begin);
    begin = end + 1;
    end = neighborstr.find_first_of(" ", begin);
    a_neighbor->eventtype = neighborstr.substr(begin, end-begin);
    begin = end + 1;
    end = neighborstr.find_first_of(" ", begin);
    a_neighbor->signaldelay = atoi(neighborstr.substr(begin, end-begin).c_str());
    begin = end + 1;
    a_neighbor->probability = atof(neighborstr.substr(begin).c_str());
    return a_neighbor;
  } */

  void sub_neighbors( std::string list ){
    if ( list.compare("") != 0 ){
      neighborhood* one_neighbor = new neighborhood();
      int begin = 0;
      int end = list.find_first_of(",", begin);
      std::string one_neighborstring;
      while ((end != string::npos) && (begin != string::npos)){
	one_neighborstring = list.substr(begin, end-begin);
	begin = list.find_first_not_of(",", end + 1);
	end = list.find_first_of(",", begin);
	one_neighbor->id = one_neighborstring;
	m_sub_neighbors.push_back(*one_neighbor);
      }
    }
  }

  /*
  void pub_neighbors( std::string list ){
    if ( list.compare("") != 0 ){
      neighborhood* one_neighbor = new neighborhood();
      int begin = 0;
      int end = list.find_first_of("|", begin);
      std::string one_neighborstring;
      while (end != string::npos){
	one_neighborstring = list.substr(begin, end-begin);
	begin = list.find_first_not_of(" ", end + 1);
	end = list.find_first_of("|", begin);
	one_neighbor = getobject(one_neighborstring);
	m_pub_neighbors.push_back(one_neighbor);
      }
    }
  }*/
  
  
  unsigned long rand_poisson_delta(unsigned long mean)
  {
    double rn = 1.0 - 1.0*rand()/(RAND_MAX+1.0);
    return (unsigned long)(-(1.0*mean)*log(rn));
  };

  int getindex( std::string type )
  {
  	if ( type.compare( "SUB" ) == 0 ){
		m_sub_index++;
		return m_sub_index;
	}
	else{
		m_pub_index++;
		return m_pub_index;
	}
  }
};

/**
 * method implementation
 */
SienaClient :: SienaClient(char* id){
  //Clear the vectors first: after the execution of this construction,
  //the process properties(including m_sub_neighbors and
  //m_pub_neighbors) will be assigned.
  m_sub_neighbors.clear();
  //m_pub_neighbors.clear();  
}

void SienaClient :: init()
{
  assert( m_unsub_timeout && m_sub_timeout && m_pub_timeout > 0.0 );

  //matches = 0;
  m_sub_index = -1;
  m_pub_index = -1;

  message_generator = new Message_Generator();
  message_generator->constr_min(m_constr_min);
  message_generator->constr_max(m_constr_max);
  message_generator->attr_min(m_attr_min);
  message_generator->attr_max(m_attr_max);
  message_generator->attr_dist_f(m_attr_dist_f);
  message_generator->constr_dist_f(m_constr_dist_f);
  message_generator->types_dist_f(m_types_dist_f);
  message_generator->op_dist_f(m_op_dist_f);
  message_generator->op_string_dist_f(m_op_string_dist_f);
  message_generator->str_dist_f(m_str_dist_f);
  message_generator->int_dist_f(m_int_dist_f);
  message_generator->bool_dist_f(m_bool_dist_f);
  message_generator->readdist();

  //schedule the first regular action
  //if ( m_left_subs > 0 ){
    //m_left_subs--;
    WeevilGen::signal_event( id(), new Trigger( id(), (std::string)"SUB", m_sub_index ) , 100); 
    //TOF: changed to m_pub_timeout to subscribe 1st time at publication time
  //}

  //if ( m_left_pubs > 0 ){
    //m_left_pubs--;
    WeevilGen::signal_event( id(), new Trigger( id(), (std::string)"PUB", m_pub_index ) , (int) (rand_poisson_delta((unsigned long) m_pub_timeout))); 
    //added some randomness
  //}
}

void SienaClient :: process_event( const Event* e )
{
  const char* msg;
  const MsgTrigger* req = dynamic_cast<const MsgTrigger*>( e );
  if(req){
      //here we got a MsgTrigger, we need to check for matching filters
      for(int i = 0; i < filters.size(); i++){
          if (filters[i].covers(req->msg)){
              ostringstream str_stream;
              //get the source of this message
              std::string from = req->from;
              str_stream << "MATCH(" <<  from << "," << req->index << ",MSG[" << req->msg_string << "],FILTER[" << filter_strings[i] << "])";
              workload_output( str_stream.str().c_str(), "", "");
              //matches++;
              return;
          }
      }
      
  } else {  
      // receive a message  
      const Trigger* rec = dynamic_cast<const Trigger*>( e );
      //get the source of this message
      std::string from = rec->from;
      //get the type of this message
      string type = rec->type;
      int index = rec->index;
    
      ostringstream str_stream;
      if (type.compare( "UNS" ) == 0){
        //this message is to schedule an unsubscription
        str_stream << "`" << type << "(" << index << ", `" + filter_strings[0] + "')'";
         //accomplish(output workload) the current scheduled action
         string str_msg = str_stream.str();
         msg = str_msg.c_str();
         workload_output( msg, "", "");
         // remove from local subscriptions
         filters.erase(filters.begin());
         filter_strings.erase(filter_strings.begin());
      }
      else if (type.compare( "SUB" ) == 0){
        //this message is to schedule a subscription
        string gen_mesg = message_generator->generate_predicate(type, &filters);        
        str_stream << "`" << type << "(" << getindex(type) << ", `\"" << gen_mesg << "\"')'";
         //accomplish(output workload) the current scheduled action
         string str_msg = str_stream.str();
         msg = str_msg.c_str();
         filter_strings.push_back(msg);
         workload_output( msg, "", "");
         //schedule the next sub
          //if ( m_left_subs > 0 ){
          //m_left_subs--;
        WeevilGen::signal_event( id(), new Trigger( id(), (std::string)"SUB", m_sub_index ) , rand_poisson_delta((unsigned long) m_sub_timeout));
          //}
          //schedule the corresponding unsub
        WeevilGen::signal_event( id(), new Trigger( id(), (std::string)"UNS", m_sub_index ), m_unsub_timeout);       
      }
       else if (type.compare( "PUB" ) == 0){
        //this message is to schedule a notification
        string gen_mesg = message_generator->generate_message(type, &messages);
        ostringstream msg_id_stream;
        msg_id_stream << id() << "m" << (m_pub_index + 1);
        string msg_id = msg_id_stream.str().c_str();
        string attrName = "msgId";
        //add unique identifier as attribute to msg                 
        messages.back().add( message_generator->new_sx_string(& attrName), new simple_value(message_generator->new_sx_string(& msg_id)));
        str_stream << "`" << type << "(" << getindex(type) << ", `\"" << gen_mesg << " & " << "s " << attrName << ' ' << msg_id << "\"')'";
         //accomplish(output workload) the current scheduled action
         string str_msg = str_stream.str();
         msg = str_msg.c_str();         
         workload_output( msg, "", "");
         // signal all neighbors (including itself) the publication of the message
          for (int i = 0; i < m_sub_neighbors.size(); i++){              
              WeevilGen::signal_event( m_sub_neighbors[i].id, new MsgTrigger(id(), messages.at(m_pub_index), msg, m_pub_index), 1);              
          }
         //schedule the next pub
         //if ( m_left_pubs > 0 ){
         //m_left_pubs--;
        WeevilGen::signal_event( id(), new Trigger( id(), (std::string)"PUB", m_pub_index ) , rand_poisson_delta((unsigned long)m_pub_timeout));
         //}                    
          
      }
  }
}

/*void SienaClient :: stop(void) {
    ostringstream str_stream;    
    str_stream << "TOTAL_MATCH(" <<  matches << ")";
    workload_output( str_stream.str().c_str(), "", "");
}*/
